library digital_3g_common;

import 'dart:convert';

import 'package:http/http.dart' as http;
import 'package:logging/logging.dart';

class TicketResponse {
  final String? ticket;
  final VerificationErrorType? error;

  TicketResponse(this.ticket, {this.error});

  bool get hasError => error != null;

  Map get toMap => {'ticket': ticket, 'error': error.toString()};

  @override
  String toString() => '[TicketResponse $toMap]';
}

class VerificationApi {
  final Logger _log = Logger('VerificationApi');

  final http.Client _client;
  final String baseUrl;

  VerificationApi(this._client, this.baseUrl);

  Future<VerificationResponse> check(String data) async {
    final response = await _client.post(
      Uri.parse('$baseUrl/check'),
      body: json.encode(data),
      headers: {
        'Content-Type': 'application/json',
      },
    );
    _log.finer('check received ${response.statusCode} status');
    if (response.statusCode == 200) {
      _log.finest(response.body);
      return VerificationResponse.fromMapNew(json.decode(response.body));
    } else if (response.statusCode == 404) {
      _log.fine('no color available');
      _log.finest(response.body);
      return VerificationResponse.fromMapNew(json.decode(response.body));
    } else if (response.statusCode >= 500) {
      _log.severe(response.body);
      return VerificationResponse.error(VerificationErrorType.server);
    } else if (response.statusCode >= 400) {
      _log.severe(response.body);
      if (response.body.contains('signing')) {
        return VerificationResponse.error(VerificationErrorType.signing);
      } else if (response.body.contains('encryption')) {
        return VerificationResponse.error(VerificationErrorType.encryption);
      }
    }
    _log.severe(response.body);
    return VerificationResponse.error(VerificationErrorType.other);
  }

  Future<TicketResponse> generateTicket(String seed) async {
    final response = await _client.post(
      Uri.parse('$baseUrl/genticket'),
      body: json.encode(seed),
      headers: {
        'Content-Type': 'application/json',
      },
    );
    _log.finer('generateTicket received ${response.statusCode} status');

    if (response.statusCode == 200) {
      _log.finest(response.body);
      return TicketResponse(json.decode(response.body));
    } else if (response.statusCode >= 500) {
      _log.severe(response.body);
      return TicketResponse(null, error: VerificationErrorType.server);
    } else if (response.statusCode >= 400) {
      _log.severe(response.body);
      if (response.body.contains('oldseed')) {
        return TicketResponse(null, error: VerificationErrorType.oldseed);
      } else if (response.body.contains('signing')) {
        return TicketResponse(null, error: VerificationErrorType.signing);
      }
    }
    _log.severe(response.body);
    return TicketResponse(null, error: VerificationErrorType.other);
  }

  Future<String> getMotd() async {
    final response = await _client.get(Uri.parse('$baseUrl/motd'));
    _log.finer('getMotd received ${response.statusCode} status');

    if (response.statusCode == 200) {
      _log.finest(response.body);
      return response.body;
    } else {
      _log.severe(response.body);
      return '';
    }
  }
}

enum VerificationErrorType {
  signing,
  encryption,
  expired,
  oldseed,
  server,
  root,
  other,
}

class VerificationResponse {
  final bool verified;
  final int? color;
  final bool? inverse;

  final DateTime timestamp;

  final VerificationErrorType? error;

  VerificationResponse(
      {required this.verified, this.color, this.inverse, required this.timestamp, this.error});

  factory VerificationResponse.empty() =>
      VerificationResponse(verified: false, timestamp: DateTime.now());

  factory VerificationResponse.error(VerificationErrorType error) => VerificationResponse(
        verified: false,
        error: error,
        timestamp: DateTime.now(),
      );

  factory VerificationResponse.fromMap(Map map) => VerificationResponse(
        verified: map['verified'],
        color: map['color'] != null ? int.tryParse(map['color'], radix: 16) : null,
        inverse: map['inverse'],
        timestamp: DateTime.tryParse(map['timestamp']) ?? DateTime(1971),
      );

  factory VerificationResponse.fromMapNew(Map map) => VerificationResponse(
        verified: map['verified'],
        color: map['color'] != null ? int.tryParse(map['color'], radix: 16) : null,
        inverse: map['inverse'],
        timestamp: DateTime.now(),
      );

  bool get hasError => error != null;

  Map<String, dynamic> get toMap => {
        'verified': verified,
        'color': color?.toRadixString(16),
        'inverse': inverse,
        'timestamp': timestamp.toIso8601String(),
        'error': error.toString(),
      };

  @override
  String toString() => '[VerificationResponse $toMap]';
}
