import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/theme.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:provider/provider.dart';

import '../api/admin_api.dart';
import '../api/auth_api.dart';
import '../models/access_code.dart';
import '../widgets/card_list_tile.dart';
import '../widgets/code_dialog.dart';
import '../widgets/csv_dialog.dart';
import '../widgets/delete_dialog.dart';

class AccessCodesPage extends StatefulWidget {
  const AccessCodesPage({Key? key}) : super(key: key);

  @override
  _AccessCodesPageState createState() => _AccessCodesPageState();
}

class _AccessCodesPageState extends State<AccessCodesPage> {
  AdminApi? _adminApi;
  List<AccessCode> _codes = [];
  List<int> _selected = [];

  bool _allSelected = false;
  bool _searchActive = false;

  String _searchTerm = '';

  List<AccessCode> _filteredCodes = [];

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Zugangscodes'),
        actions: [
          IconButton(
            onPressed: _onToggleAll,
            icon: const Icon(Icons.select_all_rounded),
            tooltip: _allSelected ? 'Alle abwählen' : 'Alle auswählen',
          ),
          IconButton(
            onPressed: _onCreateCodes,
            icon: const Icon(Icons.add),
            tooltip: 'Zugangscodes generieren',
          ),
        ],
      ),
      floatingActionButton: _selected.isEmpty
          ? SizedBox(
              height: 94.0,
              width: 256.0,
              child: Stack(
                alignment: Alignment.bottomRight,
                children: [
                  if (_searchActive)
                    Positioned(
                      left: 0.0,
                      bottom: 0.0,
                      child: Container(
                        width: 192.0,
                        height: 64.0,
                        padding: const EdgeInsets.fromLTRB(8.0, 4.0, 8.0, 8.0),
                        decoration: BoxDecoration(
                            color: Colors.white,
                            borderRadius: const BorderRadius.all(Radius.circular(16.0)),
                            border: Border.all(width: 1.0, color: darkGrey)),
                        child: TextField(
                          onChanged: _onUpdateSearch,
                          decoration: const InputDecoration(
                            hintText: 'Suchbegriff',
                            border: InputBorder.none,
                            enabledBorder: InputBorder.none,
                            disabledBorder: InputBorder.none,
                            focusedBorder: InputBorder.none,
                          ),
                        ),
                      ),
                    ),
                  Positioned(
                    right: 0.0,
                    bottom: 4.0,
                    child: FloatingActionButton(
                      onPressed: () {
                        setState(() {
                          _searchActive = !_searchActive;
                          _searchTerm = '';
                        });
                      },
                      child: Icon(_searchActive ? Icons.search_off : Icons.search),
                    ),
                  ),
                ],
              ),
            )
          : Column(
              mainAxisSize: MainAxisSize.min,
              crossAxisAlignment: CrossAxisAlignment.end,
              children: [
                FloatingActionButton(
                  heroTag: 'export_csv_tag',
                  onPressed: _onExportCSV,
                  child: const Icon(Icons.import_export),
                ),
                Padding(
                  padding: const EdgeInsets.only(top: 8.0),
                  child: FloatingActionButton(
                    heroTag: 'delete_tag',
                    onPressed: _onDeleteSelected,
                    child: const Icon(Icons.delete_forever_rounded),
                  ),
                ),
              ],
            ),
      body: Center(
        child: ConstrainedBox(
          constraints: const BoxConstraints(maxWidth: 640.0),
          child: RefreshIndicator(
            onRefresh: _onUpdateCodes,
            child: ListView(
              children: [
                ..._filteredCodes.map<Widget>(
                  (AccessCode e) => CardListTile(
                    leading: Checkbox(
                      value: _selected.contains(e.id),
                      onChanged: (bool? value) => setState(
                          () => value ?? false ? _selected.add(e.id!) : _selected.remove(e.id)),
                    ),
                    title: Text(e.code, maxLines: 1, overflow: TextOverflow.ellipsis),
                    subtitle: Text(
                      'Rolle: ${e.scope}, Eingelöst von Nutzer ${e.userId}',
                      style: Theme.of(context).textTheme.subtitle2,
                    ),
                  ),
                ),
              ],
            ),
          ),
        ),
      ),
    );
  }

  @override
  void initState() {
    super.initState();
    _initApi();
  }

  void _initApi() async {
    final _client = Provider.of<AuthApi>(context, listen: false).client;
    _adminApi = AdminApi(_client, BlocProvider.of<BackendSwitcher>(context).state);

    _onUpdateCodes();
    super.didChangeDependencies();
  }

  void _onCreateCodes() async {
    final result = await showDialog<List<AccessCode>>(
        context: context, builder: (BuildContext context) => const CodeDialog());

    if (result == null) return;

    setState(() {
      _codes.addAll(result);
    });
  }

  void _onDeleteSelected() async {
    final result = await showDialog<bool>(
      context: context,
      builder: (BuildContext context) => const DeleteDialog(),
    );

    if (!(result ?? false)) return;

    for (var id in _selected) {
      await _adminApi!.deleteAccessCode(id);
      setState(() => _codes.removeWhere((e) => e.id == id));
    }

    setState(() {
      _selected = [];
      _searchTerm = '';
      _searchActive = false;
    });
    _onUpdateSearch(_searchTerm);
  }

  void _onExportCSV() async {
    await showDialog(
      context: context,
      builder: (BuildContext context) => CsvDialog(
        filteredCodes: _filteredCodes,
        markedCodes: _selected,
      ),
    );
  }

  void _onToggleAll() {
    final ids = _filteredCodes.map<int>((e) => e.id!).toList();
    bool temp = true;
    for (var id in ids) {
      temp = _selected.contains(id);
    }
    setState(() {
      _allSelected = temp;
    });

    if (_allSelected) {
      setState(() => _selected = []);
    } else {
      setState(() => _selected = ids);
    }
  }

  Future<void> _onUpdateCodes() async {
    _codes = ((await _adminApi!.getAccessCodes()) ?? []).where((e) => e.userId == null).toList();
    setState(() => _filteredCodes = _codes);
  }

  void _onUpdateSearch(String? value) {
    _searchTerm = value ?? '';
    _filteredCodes = _codes
        .where(
          (element) =>
              !_searchActive ||
              element.code.toUpperCase().contains(_searchTerm.toUpperCase().replaceAll(' ', '')),
        )
        .toList();
    setState(() => _filteredCodes);
  }
}
