import 'dart:io';

import 'package:digital_3g_common/ui.dart';
import 'package:flutter/material.dart';
import 'package:oauth2/oauth2.dart' as oauth2;
import 'package:path_provider/path_provider.dart';
import 'package:provider/provider.dart';

import '../api/auth_api.dart';
import '../models/auth_request.dart';
import '../widgets/scope_selector.dart';

class LoginData {
  String? username;
  String? password;
  Scope mode;

  bool remember;

  LoginData({this.mode = Scope.user, this.remember = false});
}

class LoginPage extends StatefulWidget {
  const LoginPage({Key? key}) : super(key: key);

  @override
  _LoginPageState createState() => _LoginPageState();
}

class _LoginPageState extends State<LoginPage> {
  final LoginData _form = LoginData();
  File? _credentials;

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Image.asset(
          'assets/uni_logo.png',
          package: 'digital_3g_common',
          height: 40.0,
        ),
        actions: [
          IconButton(
            onPressed: () => Navigator.of(context).pushNamed('/settings'),
            icon: const Icon(Icons.settings),
          ),
        ],
        automaticallyImplyLeading: false,
      ),
      body: Center(
        child: ConstrainedBox(
          constraints: const BoxConstraints(maxWidth: 512.0),
          child: ListView(
            children: [
              PaddedCard(
                padding: const EdgeInsets.all(16.0),
                child: Form(
                  child: Column(
                    children: [
                      ListTile(
                        title: const Text('Ich bin...'),
                        trailing: SizedBox(
                          width: 128.0,
                          child: ScopeSelector(
                            onChanged: (Scope? scope) {
                              if (scope != null) _form.mode = scope;
                            },
                            value: _form.mode,
                          ),
                        ),
                      ),
                      TextFormField(
                        decoration: const InputDecoration(
                          labelText: 'Username',
                          contentPadding: EdgeInsets.all(16.0),
                          alignLabelWithHint: true,
                        ),
                        onChanged: (String input) => _form.username = input,
                      ),
                      TextFormField(
                        decoration: const InputDecoration(
                          labelText: 'Passwort',
                          contentPadding: EdgeInsets.all(16.0),
                          alignLabelWithHint: true,
                        ),
                        obscureText: true,
                        onChanged: (String input) => _form.password = input,
                      ),
                      CheckboxListTile(
                        contentPadding: const EdgeInsets.symmetric(horizontal: 64.0),
                        value: _form.remember,
                        onChanged: (bool? input) => setState(() => _form.remember = input ?? false),
                        title: const Text('Login merken'),
                      ),
                      Padding(
                        padding: const EdgeInsets.all(16.0),
                        child: Row(
                          mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                          children: [
                            MaterialButton(
                              onPressed: () => Navigator.of(context).pushNamed('/register'),
                              child: const Text('Zur Registrierung'),
                            ),
                            ElevatedButton(onPressed: _onSubmit, child: const Text('Login')),
                          ],
                        ),
                      )
                    ],
                  ),
                ),
              ),
            ],
          ),
        ),
      ),
    );
  }

  @override
  void initState() {
    _loadCredentials();
    super.initState();
  }

  Future<void> _loadCredentials() async {
    final basePath = (await getApplicationDocumentsDirectory()).path;
    _credentials = await File('$basePath/credentials').create(recursive: true);
    final credentialString = await _credentials!.readAsString();

    if (credentialString.isNotEmpty) {
      var credentials = oauth2.Credentials.fromJson(credentialString);
      final api = Provider.of<AuthApi>(context, listen: false);

      api.loadClient(oauth2.Client(credentials));
      Navigator.of(context).pop();
      if (api.user!.credentials!.scopes!.contains('admin')) {
        Navigator.of(context).pushNamed('/admin');
      } else if (api.user!.credentials!.scopes!.contains('user')) {
        Navigator.of(context).pushNamed('/');
      } else {
        await File('$basePath/credentials').delete();
        Navigator.of(context).pushNamed('/login');
      }
    } else {
      await File('$basePath/credentials').delete();
    }
  }

  Future<void> _onSubmit() async {
    if (_form.username != null &&
        _form.username!.isNotEmpty &&
        _form.password != null &&
        _form.password!.isNotEmpty) {
      final api = Provider.of<AuthApi>(context, listen: false);
      try {
        final client = (await api.postAuthToken(
          AuthRequest(
            username: _form.username!,
            password: _form.password!,
            scopes: [_form.mode],
          ),
        )) as oauth2.Client;

        if (_form.remember) {
          _credentials!.writeAsString(client.credentials.toJson());
        }

        Navigator.of(context).pop();
        if (api.user!.credentials!.scopes!.contains('admin')) {
          Navigator.of(context).pushNamed('/admin');
        } else {
          Navigator.of(context).pushNamed('/');
        }
      } catch (e) {
        ScaffoldMessenger.of(context)
            .showSnackBar(const SnackBar(content: Text('Anmeldung fehlgeschlagen')));
      }
    }
  }
}
