import 'dart:convert';

import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/content_dialog.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:provider/provider.dart';

import '../api/admin_api.dart';
import '../api/auth_api.dart';
import '../models/runtime_setting.dart';
import '../widgets/card_list_tile.dart';

const List<String> settingNames = [
  'runtime_recovered',
  'runtime_tested',
  'runtime_vaccinated',
  'min_age_recovered',
  'motd_public',
  'motd_internal',
];

const Map<String, String> settingTitles = {
  'runtime_recovered': 'Nachweislaufzeit (Genesen)',
  'runtime_tested': 'Nachweislaufzeit (Getestet)',
  'runtime_vaccinated': 'Nachweislaufzeit (Geimpft)',
  'min_age_recovered': 'Mindestalter Genesenenbescheinigung',
  'motd_public': 'Infotext (Öffentlich)',
  'motd_internal': 'Infotext (Intern)',
};

final Map<String, dynamic> settingDefaults = {
  'runtime_recovered': RuntimeSetting(RuntimeType.duration, duration: const Duration(days: 90)),
  'runtime_tested': RuntimeSetting(RuntimeType.duration, duration: const Duration(hours: 24)),
  'runtime_vaccinated': RuntimeSetting(RuntimeType.nextSemester),
  'min_age_recovered': 31,
  'motd_public': '',
  'motd_internal': '',
};

class ServerSettingsPage extends StatefulWidget {
  const ServerSettingsPage({Key? key}) : super(key: key);

  @override
  _ServerSettingsPageState createState() => _ServerSettingsPageState();
}

class SettingDialog extends StatefulWidget {
  final MapEntry setting;

  const SettingDialog(this.setting, {Key? key}) : super(key: key);

  @override
  _SettingDialogState createState() => _SettingDialogState();
}

enum TimeChoice { hours, days }

class _ServerSettingsPageState extends State<ServerSettingsPage> {
  Map<String, dynamic> _settings = {};

  AdminApi? _adminApi;

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Server-Einstellungen'),
      ),
      body: ListView(
        children: [
          ..._settings.keys.map(
            (String key) => CardListTile(
              leading: Icon(key.contains('runtime')
                  ? Icons.lock_clock_rounded
                  : key.contains('motd')
                      ? Icons.text_snippet_rounded
                      : Icons.pin_rounded),
              title: Text(settingTitles[key] ?? ''),
              onTap: () => _onTappedSetting(key, _settings[key]),
            ),
          ),
        ],
      ),
    );
  }

  @override
  void initState() {
    super.initState();
    _initApi();
  }

  Future<void> _getSettings() async {
    _settings = (await _adminApi!.getSettings());

    final missing = settingNames.where((key) => !_settings.containsKey(key));
    if (missing.isNotEmpty) {
      for (var setting in missing) {
        _settings.addAll(await _adminApi!.addSetting(setting, settingDefaults[setting].toString()));
        setState(() => _settings);
      }
    }

    for (var setting in _settings.keys) {
      if (setting.contains('runtime')) {
        _settings[setting] = RuntimeSetting.fromMap(json.decode(_settings[setting]));
      }
      if (setting.contains('min_age')) {
        _settings[setting] = int.tryParse(_settings[setting]);
      }
    }

    setState(() => _settings);
  }

  void _initApi() {
    final _client = Provider.of<AuthApi>(context, listen: false).client;
    _adminApi = AdminApi(_client, BlocProvider.of<BackendSwitcher>(context).state);
    _getSettings();
  }

  Future<void> _onTappedSetting(String key, dynamic value) async {
    final result = await showDialog(
      context: context,
      builder: (BuildContext context) => SettingDialog(MapEntry(key, value)),
    );

    if (result == null) {
      return;
    } else {
      if (result is String) {
        final response = await _adminApi!.updateSetting(key, result);
        setState(() => _settings[key] = response.values.single);
      } else if (result is RuntimeSetting) {
        final response = await _adminApi!.updateSetting(key, result.toString());
        _settings[key] = RuntimeSetting.fromMap(json.decode(response.values.single));
      } else if (result is int) {
        final response = await _adminApi!.updateSetting(key, result.toString());
        setState(() => _settings[key] = int.tryParse(response.values.single));
      }
    }
  }
}

class _SettingDialogState extends State<SettingDialog> {
  String? text;

  TimeChoice timeChoice = TimeChoice.hours;

  @override
  Widget build(BuildContext context) {
    if (widget.setting.value is RuntimeSetting) {
      final runtime = widget.setting.value as RuntimeSetting;
      if ((runtime.duration?.inHours ?? 0) > 24) timeChoice = TimeChoice.days;

      return ContentDialog(
        title: const Text('Laufzeit bearbeiten'),
        children: [
          Text(
            settingTitles[widget.setting.key] ?? '',
            style: Theme.of(context).textTheme.subtitle2,
          ),
          ListTile(
            title: const Text('Art der Laufzeit'),
            trailing: DropdownButton<RuntimeType>(
              value: runtime.type,
              items: const [
                DropdownMenuItem<RuntimeType>(
                  value: RuntimeType.duration,
                  child: Text('Tage/Stunden'),
                ),
                DropdownMenuItem<RuntimeType>(
                  value: RuntimeType.months,
                  child: Text('Monate'),
                ),
                DropdownMenuItem<RuntimeType>(
                  value: RuntimeType.nextSemester,
                  child: Text('Semesterende'),
                ),
              ],
              onChanged: (RuntimeType? value) {
                if (value != null) {
                  setState(() => runtime.type = value);
                }
              },
            ),
          ),
          if (runtime.type == RuntimeType.duration)
            ListTile(
              title: const Text('Tage/Stunden'),
              trailing: Row(
                mainAxisSize: MainAxisSize.min,
                crossAxisAlignment: CrossAxisAlignment.start,
                children: [
                  SizedBox(
                    width: 64.0,
                    child: TextFormField(
                      initialValue: timeChoice == TimeChoice.days
                          ? (runtime.duration ?? const Duration()).inDays.toString()
                          : (runtime.duration ?? const Duration()).inHours.toString(),
                      onChanged: (String input) {
                        final result = int.tryParse(input);
                        if (result == null) return;
                        setState(() => runtime.duration = timeChoice == TimeChoice.days
                            ? Duration(days: result)
                            : Duration(hours: result));
                      },
                    ),
                  ),
                  DropdownButton<TimeChoice>(
                    value: timeChoice,
                    items: const [
                      DropdownMenuItem(value: TimeChoice.hours, child: Text('Stunden')),
                      DropdownMenuItem(value: TimeChoice.days, child: Text('Tage')),
                    ],
                    onChanged: (TimeChoice? choice) =>
                        setState(() => timeChoice = choice ?? timeChoice),
                  )
                ],
              ),
            )
          else if (runtime.type == RuntimeType.months)
            ListTile(
              title: const Text('Zeitraum in Monaten'),
              trailing: SizedBox(
                width: 64.0,
                child: TextFormField(
                  initialValue: (runtime.months ?? 0).toString(),
                  onChanged: (String input) {
                    final result = int.tryParse(input);
                    if (result == null) return;
                    setState(() => runtime.months = result);
                  },
                ),
              ),
            )
        ],
        actions: [
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(), child: const Text('Abbrechen')),
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(runtime), child: const Text('Speichern')),
        ],
      );
    } else if (widget.setting.value is String) {
      return ContentDialog(
        title: const Text('Text bearbeiten'),
        children: [
          Text(
            settingTitles[widget.setting.key] ?? '',
            style: Theme.of(context).textTheme.subtitle2,
          ),
          TextFormField(
            maxLines: 10,
            initialValue: text,
            onChanged: (String input) => setState(() => text = input.trim()),
          )
        ],
        actions: [
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(), child: const Text('Abbrechen')),
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(text), child: const Text('Speichern')),
        ],
      );
    } else if (widget.setting.value is int) {
      int? days = widget.setting.value as int;
      return ContentDialog(
        title: const Text('Nummer bearbeiten'),
        children: [
          Text(
            settingTitles[widget.setting.key] ?? '',
            style: Theme.of(context).textTheme.subtitle2,
          ),
          ListTile(
            title: const Text('Tage'),
            trailing: SizedBox(
              width: 128.0,
              child: TextFormField(
                initialValue: days.toString(),
                keyboardType: TextInputType.number,
                onChanged: (String input) => days = int.tryParse(input),
              ),
            ),
          ),
        ],
        actions: [
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(), child: const Text('Abbrechen')),
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(days), child: const Text('Speichern')),
        ],
      );
    } else {
      return ContentDialog(
        title: const Text('Unbekannt bearbeiten'),
        children: [
          TextFormField(
            maxLines: 10,
            initialValue: text,
            onChanged: (String input) => setState(() => text = input),
          )
        ],
        actions: [
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(), child: const Text('Abbrechen')),
          ElevatedButton(
              onPressed: () => Navigator.of(context).pop(text), child: const Text('Speichern')),
        ],
      );
    }
  }

  @override
  void initState() {
    super.initState();
    text = widget.setting.value is String ? widget.setting.value : null;
  }
}
