import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:bloc/bloc.dart';
import 'package:connectivity/connectivity.dart';
import 'package:crypto/crypto.dart';
import 'package:digital_3g_common/verification_api.dart';
import 'package:encrypt/encrypt.dart';
import 'package:flutter_jailbreak_detection/flutter_jailbreak_detection.dart';
import 'package:http/http.dart' as http;
import 'package:logging/logging.dart';
import 'package:path_provider/path_provider.dart';
import 'package:platform_device_id/platform_device_id.dart';

const String iv = 'c3eb69edb728ff3c';

enum EventType {
  check,
  revoke,
  add,
}

class VerificationBloc extends Bloc<VerificationEvent, VerificationState> {
  final Logger _log = Logger('VerificationBloc');

  File? _ticketStore;
  String? _ticket;

  File? _verificationStore;

  VerificationApi? _api;
  VerificationResponse? _verification;
  VerificationErrorType? _error;

  Encrypter? _encrypter;

  VerificationBloc(String url) : super(VerificationState.busy) {
    _api = VerificationApi(http.Client(), url);
    FlutterJailbreakDetection.jailbroken.then((bool jailbroken) {
      if (!jailbroken) {
        _initVerificationStore()
            .then((_) => _initTicketFile().then((_) => _getDeviceId().then((_) => _loadTicket())));
      } else {
        _log.severe('jailbroken/rooted device detected, VerificationBloc halted');
        _error = VerificationErrorType.root;
        emit(VerificationState.error);
      }
    });
  }

  VerificationApi? get api => _api;
  VerificationErrorType? get error => _error;
  String? get ticket => _ticket;
  VerificationResponse? get verification => _verification;

  @override
  Stream<VerificationState> mapEventToState(VerificationEvent event) async* {
    _log.fine('got event $event');
    switch (event.type) {
      case EventType.check:
        if (_ticket != null && _ticket!.isNotEmpty) {
          VerificationResponse? response;
          try {
            response = await _api!.check(_ticket!);
            _log.finer(response);
          } catch (e) {
            _log.severe('failed to check ticket:\n$e');
          }
          if ((response?.verified ?? false) && !(response?.hasError ?? false)) {
            _verification = response;
            final iv = IV.fromLength(16);
            await _verificationStore!.writeAsString(
                iv.base64 + _encrypter!.encrypt(json.encode(_verification!.toMap), iv: iv).base64);
            yield VerificationState.verified;
          } else if (((((await Connectivity().checkConnectivity()) == ConnectivityResult.none) ||
                  ((response?.hasError ?? false) &&
                      response!.error == VerificationErrorType.server)) &&
              (_verificationStore?.existsSync() ?? false))) {
            // get verification from local storage
            String rawVerification;
            final storeData = await _verificationStore!.readAsString();
            final iv = IV.fromBase64(storeData.substring(0, 24));
            final encrypted = Encrypted.fromBase64(storeData.substring(24));
            try {
              rawVerification = _encrypter!.decrypt(encrypted, iv: iv);
            } catch (e) {
              yield VerificationState.unverified;
              return;
            }
            final loaded = VerificationResponse.fromMap(json.decode(rawVerification));
            if (loaded.timestamp.isAfter(DateTime.now().subtract(const Duration(minutes: 30)))) {
              _verification = loaded;
              yield VerificationState.verified;
            } else {
              yield VerificationState.unverified;
            }
          } else if (!(response?.verified ?? false) && !(response?.hasError ?? false)) {
            yield VerificationState.unverified;
          } else {
            _verification = response;
            _error = _verification!.error;
            yield VerificationState.error;
          }
        }
        break;
      case EventType.revoke:
        _ticketStore!.writeAsString('', flush: true);
        _ticket = null;
        _error = null;
        yield VerificationState.noTicket;
        break;
      case EventType.add:
        // get ticket from backend if longer than 728 (backwards compatibility)
        TicketResponse? response;
        if ((event.seed?.length ?? 0) > 768) {
          try {
            response = await _api!.generateTicket(event.seed!);
          } catch (e) {
            _log.severe('error while generating ticket');
            _error = VerificationErrorType.other;
            yield VerificationState.error;
          }
        } else {
          response = TicketResponse(event.seed);
        }

        if (response!.hasError) {
          _error = response.error;
          yield VerificationState.error;
        }

        // save ticket
        if (!_ticketStore!.existsSync()) {
          await _ticketStore!.create(recursive: true);
        }
        if (response.ticket != null && response.ticket!.isNotEmpty) {
          _ticket = response.ticket;
          _writeEncryptedTicket(_ticket!);
          yield VerificationState.unverified;
          add(VerificationEvent.check());
        }
        break;
      default:
        return;
    }
  }

  void onAdd(String ticket) => add(VerificationEvent.add(ticket));

  @override
  void onChange(Change<VerificationState> change) {
    _log.finer('new state: ${change.nextState}');
    super.onChange(change);
  }

  void onCheck() => add(VerificationEvent.check());

  void onRevoke() => add(VerificationEvent.revoke());

  Future<void> _getDeviceId() async {
    final deviceId = await PlatformDeviceId.getDeviceId;
    final key = Key(sha256.convert(utf8.encode(deviceId!)).bytes as Uint8List);

    _encrypter = Encrypter(AES(key, mode: AESMode.cbc));
  }

  Future<void> _initTicketFile() async {
    _ticketStore = File('${(await getApplicationDocumentsDirectory()).path}/ticket');
    _log.fine('initialized ticket store in ${_ticketStore?.path}');
  }

  Future<void> _initVerificationStore() async {
    _verificationStore = File('${(await getApplicationDocumentsDirectory()).path}/last');
    _log.fine('initialized verification store in ${_verificationStore?.path}');
  }

  Future<void> _loadTicket() async {
    String? ticket;
    if (!_ticketStore!.existsSync()) {
      _log.fine('no ticket found, creating ticket store');
      _ticketStore!.create(recursive: true);
      add(VerificationEvent.revoke());
      return;
    }

    _log.fine('ticketStore found, loading and checking');
    ticket = await _ticketStore!.readAsString();
    _log.finer('ticket: "$ticket"');
    if (ticket.isNotEmpty) {
      final iv = IV.fromBase64(ticket.substring(0, 24));
      final encrypted = Encrypted.fromBase64(ticket.substring(24));
      try {
        _ticket = _encrypter!.decrypt(encrypted, iv: iv);
      } catch (e) {
        // convert old ticket into encrypted
        _writeEncryptedTicket(ticket);
        _log.info('restarting loadTicket with migrated ticket');
        _loadTicket();
        return;
      }

      add(VerificationEvent.check());
    } else {
      _log.fine('ticket empty, going to noTicket state');
      add(VerificationEvent.revoke());
    }
  }

  Future<void> _writeEncryptedTicket(String ticket) async {
    final iv = IV.fromLength(16);
    // write 24 characters of IV and append encrypted data to ticketStore
    await _ticketStore!.writeAsString(
      iv.base64 + _encrypter!.encrypt(ticket, iv: iv).base64,
      flush: true,
    );
  }
}

class VerificationEvent {
  final EventType type;
  final String? seed;

  VerificationEvent({required this.type, this.seed});

  factory VerificationEvent.add(String ticket) =>
      VerificationEvent(type: EventType.add, seed: ticket);

  factory VerificationEvent.check() => VerificationEvent(type: EventType.check);

  factory VerificationEvent.revoke() => VerificationEvent(type: EventType.revoke);

  Map get toMap => {'type': type, 'seed': seed};

  @override
  String toString() => '[VerificationEvent $toMap]';
}

enum VerificationState {
  noTicket,
  verified,
  unverified,
  error,
  busy,
}
